package com.jdbernard.timestamper.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Represents a Timeline configuration. A configuration has one primary, or
 * local, Timeline with an associated TimelineSource, and a list of remote
 * Timelines synched to the local Timeline.
 * <br>
 * <table>
 * <tr><th>Property</th><th>Description</th></tr>
 * <tr><td><code>timeline.uri</code></td><td>The URI of the primary (local)
 *      timeline</td></tr>
 * <tr><td><code>remote.timeline.</code><i>name</i><code>.uri</code></td>
 *      <td>The URI for the <i>name</i> remote timeline.</td></tr>
 * <tr><td><code>remote.timeline.</code><i>name</i><code>.push</code></td>
 *      <td><code>true</code> to enable pushing updates to the <i>name</i>
 *      remote timeline, <code>false</code> to disable.</td></tr>
 * <tr><td><code>remote.timeline.</code><i>name</i><code>.pull</code></td>
 *      <td><code>true</code> to enable pulling updates from the <i>name</i>
 *      remote timeline, <code>false</code> to disable.</td></tr>
 * <tr><td><code>remote.timeline.</code><i>name</i><code>.save-on-exit</code>
 *      </td><td><code>true</code> to force sync the <i>name</i> remote
 *      timeline on exit.</td></tr>
 * <tr><td><code>remote.timeline.</code><i>name</i>
 *      <code>.update-interval</code></td><td>The time in milliseconds between
 *      synching the <i>name</i> remote timeline.</td></tr></table>
 * @author Jonathan Bernard ({@literal jonathan.bernard@gemalto.com})
 * @see com.jdbernard.timestamper.core.Timeline
 * @see com.jdbernard.timestamper.core.TimelineSource
 * @see com.jdbernard.timestamper.core.SyncTarget
 */
public class TimelineProperties {

    public static final String LOCAL_TIMELINE_URI = "timeline.uri";
    public static final String REMOTE_TIMELINE_BASE = "remote.timeline.";

    private static final Pattern remoteTimelinePropPattern = 
            Pattern.compile("\\Q" + REMOTE_TIMELINE_BASE + "\\E([^\\s\\.=]+?)[\\.=].*");

    private Timeline timeline;
    private TimelineSource timelineSource;
    private LinkedList<SyncTarget> syncTargets = new LinkedList<SyncTarget>();

    /**
     * Create new TimelineProperties, using default values. This will create
     * a new configuration using a FileTimelineSource pointed at
     * <code>'timeline.default.txt'</code> in the current directory and no
     * remote Timelines. It will save this configuration to 
     * <code>'timeline.default.properties'</code> in the current directory.
     */
    public TimelineProperties() {
        File propertyFile = new File("timeline.default.properties");
        Properties config = new Properties();

        File timelineFile = new File("timeline.default.txt");
        URI timelineURI = timelineFile.toURI();

        timeline = new Timeline();
        timelineSource = TimelineSourceFactory.newInstance(timelineURI);
        try { timelineSource.persist(timeline); }
        catch (IOException ioe) {
            // TODO
        }

        config.setProperty(LOCAL_TIMELINE_URI, timelineURI.toString());

        try { config.store(new FileOutputStream(propertyFile), ""); }
        catch (IOException ioe) {
            // TODO
        }
    }

    /**
     * Load TimelineProperties from an InputStream.
     * @param is
     */
    public TimelineProperties(InputStream is) throws IOException {
        String strURI;
        URI timelineURI;

        Properties config = new Properties();
        try {
            config.load(new InputStreamReader(is));
        } catch (IOException ioe) {
            // TODO
        }

        // load local timeline
        strURI = config.getProperty(LOCAL_TIMELINE_URI, "");
        if ("".equals(strURI)) {
            File defaultTimelineFile = new File("timeline.default.txt");
            try {
                if (!defaultTimelineFile.exists())
                    defaultTimelineFile.createNewFile();
            } catch (IOException ioe) {
                // TODO
            }
            timelineURI = defaultTimelineFile.toURI();
        } else {
            try { timelineURI = new URI(strURI); }
            catch (URISyntaxException urise) {
                throw new IOException("Unable to load the timeline: the timeline "
                        + "URI is invalid.", urise);
            }
        }

        timelineSource = TimelineSourceFactory.newInstance(timelineURI);
        timeline = timelineSource.read();

        // search keys for remote timeline entries
        for (Object keyObj : config.keySet()) {
            if (!(keyObj instanceof String)) continue;

            String key = (String) keyObj;
            String stName;
            String remoteBase;
            SyncTarget st;

            Matcher m = remoteTimelinePropPattern.matcher(key);
            if (!m.matches()) continue;

            stName = m.group(1);
            remoteBase = REMOTE_TIMELINE_BASE + stName;

            strURI = config.getProperty(remoteBase + ".uri", "");
            try { timelineURI = new URI(strURI); }
            catch (URISyntaxException urise) { /* TODO */ }

            // add a new SyncTarget to the list
            st = new SyncTarget(stName, TimelineSourceFactory
                    .newInstance(timelineURI), timeline);
            syncTargets.add(st);

            // check for synch options
            st.setPullEnabled(Boolean.parseBoolean(
                    config.getProperty(remoteBase + ".pull", "true")));
            st.setPushEnabled(Boolean.parseBoolean(
                    config.getProperty(remoteBase + ".push", "true")));
            st.setSyncOnExit(Boolean.parseBoolean(
                    config.getProperty(remoteBase + ".sync-on-exit", "true")));
            st.setSyncInterval(Long.parseLong(
                    config.getProperty(remoteBase + ".update-interval",
                    "1800000"))); // thirty minutes

        }
    }

    public void save(OutputStream os) throws IOException {
        Properties config = new Properties();
        timelineSource.persist(timeline);

        config.setProperty(LOCAL_TIMELINE_URI,
                timelineSource.getURI().toString());

        for (SyncTarget st : syncTargets) {
            String remoteBase = REMOTE_TIMELINE_BASE + st.getName();
            config.setProperty(remoteBase + ".uri",
                    st.getSource().getURI().toString());
            config.setProperty(remoteBase + ".pull",
                    Boolean.toString(st.getPullEnabled()));
            config.setProperty(remoteBase + ".push",
                    Boolean.toString(st.getPushEnabled()));
            config.setProperty(remoteBase + ".sync-on-exit",
                    Boolean.toString(st.getSyncOnExit()));
            config.setProperty(remoteBase + ".update-interval",
                    Long.toString(st.getSyncInterval()));
        }

        try {
            config.store(os, "");
        } catch (IOException ioe) {
            // TODO
        }
    }

    public Timeline getTimeline() { return timeline; }

    public void setTimelineSource(TimelineSource newSource) {
        this.timelineSource = newSource;
    }

    public TimelineSource getTimelineSource() { return timelineSource; }

    public Collection<SyncTarget> getSyncTargets() { return syncTargets; }
}
